if (Symbol && !Symbol.asyncIterator) {
    Symbol.asyncIterator = Symbol.for("__@@asyncIterator__");
}
class BatchOperation {
    /**
     * @param client    The AWS SDK client with which to communicate with
     *                  DynamoDB.
     * @param items     A synchronous or asynchronous iterable of tuples
     *                  describing the operations to execute. The first member
     *                  of the tuple should be the name of the table targeted by
     *                  the operation.
     */
    constructor(client, items) {
        this.client = client;
        /**
         * Items that have been retrieved and are ready to be returned.
         */
        this.pending = [];
        /**
         * A mapping of table names to table-specific operation state (e.g., the
         * number of throttling events experienced, etc.)
         */
        this.state = {};
        /**
         * Input elements that are prepared for immediate dispatch
         */
        this.toSend = [];
        this.throttled = new Set();
        this.sourceDone = false;
        if (isIterable(items)) {
            this.iterator = items[Symbol.iterator]();
        }
        else {
            this.iterator = items[Symbol.asyncIterator]();
        }
        this.sourceNext = this.iterator.next();
    }
    next() {
        if (this.lastResolved) {
            this.lastResolved = this.lastResolved.then(() => this.getNext());
        }
        else {
            this.lastResolved = this.getNext();
        }
        return this.lastResolved;
    }
    [Symbol.asyncIterator]() {
        return this;
    }
    /**
     * Create and return the initial state object for a given DynamoDB table.
     *
     * @param tableName The name of the table whose initial state should be
     *                  returned.
     */
    getInitialTableState(tableName) {
        return {
            backoffFactor: 0,
            name: tableName,
        };
    }
    /**
     * Accept an array of unprocessed items belonging to a single table and
     * re-enqueue it for submission, making sure the appropriate level of
     * backoff is applied to future operations on the same table.
     *
     * @param tableName     The table to which the unprocessed elements belong.
     * @param unprocessed   Elements returned by DynamoDB as not yet processed.
     *                      The elements should not be unmarshalled, but they
     *                      should be reverted to the form used for elements
     *                      that have not yet been sent.
     */
    handleThrottled(tableName, unprocessed) {
        const tableState = this.state[tableName];
        tableState.backoffFactor++;
        if (tableState.tableThrottling) {
            this.throttled.delete(tableState.tableThrottling.backoffWaiter);
            unprocessed.unshift(...tableState.tableThrottling.unprocessed);
        }
        tableState.tableThrottling = {
            unprocessed,
            backoffWaiter: new Promise(resolve => {
                setTimeout(resolve, exponentialBackoff(tableState.backoffFactor), tableState);
            })
        };
        this.throttled.add(tableState.tableThrottling.backoffWaiter);
    }
    /**
     * Iterate over all pending writes and move those targeting throttled tables
     * into the throttled queue.
     *
     * @param unprocessedTables     A set of tables for which some items were
     *                              returned without being processed.
     */
    movePendingToThrottled(unprocessedTables) {
        for (let i = this.toSend.length - 1; i > -1; i--) {
            const [table, attributes] = this.toSend[i];
            if (unprocessedTables.has(table)) {
                this.state[table]
                    .tableThrottling.unprocessed.push(attributes);
                this.toSend.splice(i, 1);
            }
        }
    }
    addToSendQueue([tableName, attributes]) {
        if (!this.state[tableName]) {
            this.state[tableName] = this.getInitialTableState(tableName);
        }
        const tableState = this.state[tableName];
        if (tableState.tableThrottling) {
            tableState.tableThrottling.unprocessed.push(attributes);
        }
        else {
            this.toSend.push([tableName, attributes]);
        }
    }
    enqueueThrottled(table) {
        const { tableThrottling: { backoffWaiter, unprocessed } } = table;
        if (unprocessed.length > 0) {
            this.toSend.push(...unprocessed.map(attr => [table.name, attr]));
        }
        this.throttled.delete(backoffWaiter);
        delete table.tableThrottling;
    }
    async getNext() {
        if (this.sourceDone &&
            this.pending.length === 0 &&
            this.toSend.length === 0 &&
            this.throttled.size === 0) {
            return { done: true };
        }
        if (this.pending.length > 0) {
            return {
                done: false,
                value: this.pending.shift()
            };
        }
        await this.refillPending();
        return this.getNext();
    }
    async refillPending() {
        while (!this.sourceDone &&
            this.toSend.length < this.batchSize) {
            const toProcess = isIteratorResult(this.sourceNext)
                ? this.sourceNext
                : await Promise.race([
                    this.sourceNext,
                    Promise.race(this.throttled)
                ]);
            if (isIteratorResult(toProcess)) {
                this.sourceDone = toProcess.done;
                if (!this.sourceDone) {
                    this.addToSendQueue(toProcess.value);
                    this.sourceNext = this.iterator.next();
                }
            }
            else {
                this.enqueueThrottled(toProcess);
            }
        }
        while (this.toSend.length < this.batchSize && this.throttled.size > 0) {
            this.enqueueThrottled(await Promise.race(this.throttled));
        }
        if (this.toSend.length > 0) {
            await this.doBatchRequest();
        }
    }
}
function exponentialBackoff(attempts) {
    return Math.floor(Math.random() * Math.pow(2, attempts));
}
function isIterable(arg) {
    return Boolean(arg) && typeof arg[Symbol.iterator] === 'function';
}
function isIteratorResult(arg) {
    return Boolean(arg) && typeof arg.done === 'boolean';
}

const MAX_READ_BATCH_SIZE = 100;
/**
 * Retrieves items from DynamoDB in batches of 100 or fewer via one or more
 * BatchGetItem operations. The items may be from any number of tables.
 *
 * This method will automatically retry any get requests returned by DynamoDB as
 * unprocessed. Exponential backoff on unprocessed items is employed on a
 * per-table basis.
 */
class BatchGet extends BatchOperation {
    /**
     * @param client    The AWS SDK client with which to communicate with
     *                  DynamoDB.
     * @param items     A synchronous or asynchronous iterable of tuples
     *                  describing the reads to execute. The first member of the
     *                  tuple should be the name of the table from which to
     *                  read, and the second should be the marshalled key.
     * @param options   Additional options to apply to the operations executed.
     */
    constructor(client, items, { ConsistentRead, PerTableOptions = {}, } = {}) {
        super(client, items);
        this.batchSize = MAX_READ_BATCH_SIZE;
        this.consistentRead = ConsistentRead;
        this.options = PerTableOptions;
    }
    async doBatchRequest() {
        const operationInput = { RequestItems: {} };
        let batchSize = 0;
        while (this.toSend.length > 0) {
            const [tableName, item] = this.toSend.shift();
            if (operationInput.RequestItems[tableName] === undefined) {
                const { projection, consistentRead, attributeNames, } = this.state[tableName];
                operationInput.RequestItems[tableName] = {
                    Keys: [],
                    ConsistentRead: consistentRead,
                    ProjectionExpression: projection,
                    ExpressionAttributeNames: attributeNames,
                };
            }
            operationInput.RequestItems[tableName].Keys.push(item);
            if (++batchSize === this.batchSize) {
                break;
            }
        }
        const { Responses = {}, UnprocessedKeys = {}, } = await this.client.batchGetItem(operationInput).promise();
        const unprocessedTables = new Set();
        for (const table of Object.keys(UnprocessedKeys)) {
            unprocessedTables.add(table);
            this.handleThrottled(table, UnprocessedKeys[table].Keys);
        }
        this.movePendingToThrottled(unprocessedTables);
        for (const table of Object.keys(Responses)) {
            const tableData = this.state[table];
            tableData.backoffFactor = Math.max(0, tableData.backoffFactor - 1);
            for (const item of Responses[table]) {
                this.pending.push([table, item]);
            }
        }
    }
    getInitialTableState(tableName) {
        const { ExpressionAttributeNames, ProjectionExpression, ConsistentRead = this.consistentRead, } = this.options[tableName] || {};
        return Object.assign({}, super.getInitialTableState(tableName), { attributeNames: ExpressionAttributeNames, projection: ProjectionExpression, consistentRead: ConsistentRead });
    }
}

const bytes = require('utf8-bytes');
/**
 * @internal
 */
function itemIdentifier(tableName, { DeleteRequest, PutRequest }) {
    if (DeleteRequest) {
        return `${tableName}::delete::${serializeKeyTypeAttributes(DeleteRequest.Key)}`;
    }
    else if (PutRequest) {
        return `${tableName}::put::${serializeKeyTypeAttributes(PutRequest.Item)}`;
    }
    throw new Error(`Invalid write request provided`);
}
function serializeKeyTypeAttributes(attributes) {
    const keyTypeProperties = [];
    for (const property of Object.keys(attributes).sort()) {
        const attribute = attributes[property];
        if (attribute.B) {
            keyTypeProperties.push(`${property}=${toByteArray(attribute.B)}`);
        }
        else if (attribute.N) {
            keyTypeProperties.push(`${property}=${attribute.N}`);
        }
        else if (attribute.S) {
            keyTypeProperties.push(`${property}=${attribute.S}`);
        }
    }
    return keyTypeProperties.join('&');
}
function toByteArray(value) {
    if (ArrayBuffer.isView(value)) {
        return new Uint8Array(value.buffer, value.byteOffset, value.byteLength);
    }
    if (typeof value === 'string') {
        return Uint8Array.from(bytes(value));
    }
    if (isArrayBuffer(value)) {
        return new Uint8Array(value);
    }
    throw new Error('Unrecognized binary type');
}
function isArrayBuffer(arg) {
    return (typeof ArrayBuffer === 'function' && arg instanceof ArrayBuffer) ||
        Object.prototype.toString.call(arg) === '[object ArrayBuffer]';
}

const MAX_WRITE_BATCH_SIZE = 25;
/**
 * Puts or deletes items from DynamoDB in batches of 25 or fewer via one or more
 * BatchWriteItem operations. The items may belong to any number of tables.
 *
 * The iterable of writes to perform may be synchronous or asynchronous and is
 * expected to yield tuples describing the writes to be performed. The first
 * member should be the table name, and the second should be {WriteRequest}
 * object that defines either a put request or a delete request.
 *
 * This method will automatically retry any write requests returned by DynamoDB
 * as unprocessed. Exponential backoff on unprocessed items is employed on a
 * per-table basis.
 */
class BatchWrite extends BatchOperation {
    constructor() {
        super(...arguments);
        this.batchSize = MAX_WRITE_BATCH_SIZE;
    }
    async doBatchRequest() {
        const inFlight = [];
        const operationInput = { RequestItems: {} };
        let batchSize = 0;
        while (this.toSend.length > 0) {
            const [tableName, marshalled] = this.toSend.shift();
            inFlight.push([tableName, marshalled]);
            if (operationInput.RequestItems[tableName] === undefined) {
                operationInput.RequestItems[tableName] = [];
            }
            operationInput.RequestItems[tableName].push(marshalled);
            if (++batchSize === this.batchSize) {
                break;
            }
        }
        const { UnprocessedItems = {} } = await this.client.batchWriteItem(operationInput).promise();
        const unprocessedTables = new Set();
        for (const table of Object.keys(UnprocessedItems)) {
            unprocessedTables.add(table);
            const unprocessed = [];
            for (const item of UnprocessedItems[table]) {
                if (item.DeleteRequest || item.PutRequest) {
                    unprocessed.push(item);
                    const identifier = itemIdentifier(table, item);
                    for (let i = inFlight.length - 1; i >= 0; i--) {
                        const [tableName, attributes] = inFlight[i];
                        if (tableName === table &&
                            itemIdentifier(tableName, attributes) === identifier) {
                            inFlight.splice(i, 1);
                        }
                    }
                }
            }
            this.handleThrottled(table, unprocessed);
        }
        this.movePendingToThrottled(unprocessedTables);
        const processedTables = new Set();
        for (const [tableName, marshalled] of inFlight) {
            processedTables.add(tableName);
            this.pending.push([tableName, marshalled]);
        }
        for (const tableName of processedTables) {
            this.state[tableName].backoffFactor =
                Math.max(0, this.state[tableName].backoffFactor - 1);
        }
    }
}

export { MAX_READ_BATCH_SIZE, BatchGet, MAX_WRITE_BATCH_SIZE, BatchWrite };
